from ._victor_plonk import _VictorPlonk

class _VictorOverridables(_VictorPlonk):

    def post_params_step(self):
        """
        This method is intended for make inherited mods easier.
        :return:
        """
        pass

    def post_monster_step(self):
        """
        This method is intended for make inherited mods easier.
        :return:
        """
        pass

    def pose_mod_step(self):
        """
        This method is intended for make inherited mods easier.
        :return:
        """
        pass

    def pre_igor_step(self):
        """
        This method is intended for make inherited mods easier.
        :return:
        """
        pass

    def post_igor_step(self):
        """
        This method is intended for make inherited mods easier.
        :return:
        """
        pass